mirror of https://lore.kernel.org/lkml/
 help / color / mirror / Atom feed
* [PATCH v10 0/2] Make find_later_rq() choose a closer cpu in topology
@ 2017-08-24  1:06 Byungchul Park
  2017-08-24  1:06 ` [PATCH v10 1/2] sched/deadline: Add support for SD_PREFER_SIBLING on find_later_rq() Byungchul Park
  2017-08-24  1:06 ` [PATCH v10 2/2] sched/rt: Add support for SD_PREFER_SIBLING on find_lowest_rq() Byungchul Park
  0 siblings, 2 replies; 7+ messages in thread
From: Byungchul Park @ 2017-08-24  1:06 UTC (permalink / raw)
  To: peterz, mingo, rostedt
  Cc: joel.opensrc, linux-kernel, juri.lelli, kernel-team

Change from v9
   -. modify a comment a bit so to be more clear

Change from v8
   -. add suggested-by Peterz
   -. add several comments

Change from v7
   -. fix a trivial typo
   -. modify commit messages to explain what it does more clearly
   -. simplify code with an existing macro

Change from v6
   -. add a comment about selection of fallback_cpu incase more than one exist
   -. modify a comment explaining what we do wrt PREFER_SIBLING

Change from v5
   -. exclude two patches already picked up by peterz
      (sched/deadline: Make find_later_rq() choose a closer cpu in topology)
      (sched/deadline: Change return value of cpudl_find())
   -. apply what peterz fixed for 'prefer sibling', into deadline and rt

Change from v4
   -. remove a patch that might cause huge lock contention
      (by spin lock(&cpudl.lock) in a hot path of scheduler)

Change from v3
   -. rename closest_cpu to best_cpu so that it align with rt
   -. protect referring cpudl.elements with cpudl.lock
   -. change return value of cpudl_find() to bool

Change from v2
   -. add support for SD_PREFER_SIBLING

Change from v1
   -. clean up the patch

Byungchul Park (2):
  sched/deadline: Add support for SD_PREFER_SIBLING on find_later_rq()
  sched/rt: Add support for SD_PREFER_SIBLING on find_lowest_rq()

 kernel/sched/deadline.c | 80 ++++++++++++++++++++++++++++++++++++++++++++-----
 kernel/sched/rt.c       | 78 +++++++++++++++++++++++++++++++++++++++++++++--
 2 files changed, 148 insertions(+), 10 deletions(-)

-- 
1.9.1

^ permalink raw reply	[flat|nested] 7+ messages in thread

* [PATCH v10 1/2] sched/deadline: Add support for SD_PREFER_SIBLING on find_later_rq()
  2017-08-24  1:06 [PATCH v10 0/2] Make find_later_rq() choose a closer cpu in topology Byungchul Park
@ 2017-08-24  1:06 ` Byungchul Park
  2017-08-24  1:06 ` [PATCH v10 2/2] sched/rt: Add support for SD_PREFER_SIBLING on find_lowest_rq() Byungchul Park
  1 sibling, 0 replies; 7+ messages in thread
From: Byungchul Park @ 2017-08-24  1:06 UTC (permalink / raw)
  To: peterz, mingo, rostedt
  Cc: joel.opensrc, linux-kernel, juri.lelli, kernel-team

It would be better to try to check other siblings first if
SD_PREFER_SIBLING is flaged when pushing tasks - migration.

Suggested-by: Peter Zijlstra <peterz@infradead.org>
Signed-off-by: Byungchul Park <byungchul.park@lge.com>
Acked-by: Juri Lelli <juri.lelli@arm.com>
---
 kernel/sched/deadline.c | 80 ++++++++++++++++++++++++++++++++++++++++++++-----
 1 file changed, 73 insertions(+), 7 deletions(-)

-- 
1.9.1

^ permalink raw reply	[flat|nested] 7+ messages in thread

* [PATCH v10 2/2] sched/rt: Add support for SD_PREFER_SIBLING on find_lowest_rq()
  2017-08-24  1:06 [PATCH v10 0/2] Make find_later_rq() choose a closer cpu in topology Byungchul Park
  2017-08-24  1:06 ` [PATCH v10 1/2] sched/deadline: Add support for SD_PREFER_SIBLING on find_later_rq() Byungchul Park
@ 2017-08-24  1:06 ` Byungchul Park
  2017-08-24 13:55   ` Steven Rostedt
  1 sibling, 1 reply; 7+ messages in thread
From: Byungchul Park @ 2017-08-24  1:06 UTC (permalink / raw)
  To: peterz, mingo, rostedt
  Cc: joel.opensrc, linux-kernel, juri.lelli, kernel-team

Hello Steven,

Can I keep your reviewed-by with comments modified?

----->8-----
>From 285e72aff363f3f84a8ccf05c7236a0c849f8944 Mon Sep 17 00:00:00 2001
From: Byungchul Park <byungchul.park@lge.com>
Date: Thu, 24 Aug 2017 09:59:10 +0900
Subject: [PATCH v10 2/2] sched/rt: Add support for SD_PREFER_SIBLING on
 find_lowest_rq()

It would be better to try to check other siblings first if
SD_PREFER_SIBLING is flaged when pushing tasks - migration.

Suggested-by: Peter Zijlstra <peterz@infradead.org>
Signed-off-by: Byungchul Park <byungchul.park@lge.com>
---
 kernel/sched/rt.c | 78 ++++++++++++++++++++++++++++++++++++++++++++++++++++---
 1 file changed, 75 insertions(+), 3 deletions(-)

diff --git a/kernel/sched/rt.c b/kernel/sched/rt.c
index 979b734..2df7813 100644
--- a/kernel/sched/rt.c
+++ b/kernel/sched/rt.c
@@ -1618,12 +1618,35 @@ static struct task_struct *pick_highest_pushable_task(struct rq *rq, int cpu)
 
 static DEFINE_PER_CPU(cpumask_var_t, local_cpu_mask);
 
+/*
+ * Find the first cpu in: mask & sd & ~prefer
+ */
+static int find_cpu(const struct cpumask *mask,
+		    const struct sched_domain *sd,
+		    const struct sched_domain *prefer)
+{
+	const struct cpumask *sds = sched_domain_span(sd);
+	const struct cpumask *ps  = prefer ? sched_domain_span(prefer) : NULL;
+	int cpu;
+
+	for_each_cpu(cpu, mask) {
+		if (!cpumask_test_cpu(cpu, sds))
+			continue;
+		if (ps && cpumask_test_cpu(cpu, ps))
+			continue;
+		break;
+	}
+
+	return cpu;
+}
+
 static int find_lowest_rq(struct task_struct *task)
 {
-	struct sched_domain *sd;
+	struct sched_domain *sd, *prefer = NULL;
 	struct cpumask *lowest_mask = this_cpu_cpumask_var_ptr(local_cpu_mask);
 	int this_cpu = smp_processor_id();
 	int cpu      = task_cpu(task);
+	int fallback_cpu = -1;
 
 	/* Make sure the mask is initialized first */
 	if (unlikely(!lowest_mask))
@@ -1668,9 +1691,35 @@ static int find_lowest_rq(struct task_struct *task)
 				return this_cpu;
 			}
 
-			best_cpu = cpumask_first_and(lowest_mask,
-						     sched_domain_span(sd));
+			/*
+			 * If a cpu being in lowest_mask & current sd &
+			 * ~prefer sd is valid, that becomes our choice.
+			 * Of course, the lowest possible cpu is already
+			 * under consideration through lowest_mask.
+			 */
+			best_cpu = find_cpu(lowest_mask, sd, prefer);
+
 			if (best_cpu < nr_cpu_ids) {
+				/*
+				 * If current domain is SD_PREFER_SIBLING
+				 * flaged, we have to try to check other
+				 * siblings first.
+				 */
+				if (sd->flags & SD_PREFER_SIBLING) {
+					prefer = sd;
+
+					/*
+					 * fallback_cpu should be one
+					 * in the closest domain among
+					 * SD_PREFER_SIBLING domains,
+					 * in case that more than one
+					 * SD_PREFER_SIBLING domains
+					 * exist in the hierachy.
+					 */
+					if (fallback_cpu == -1)
+						fallback_cpu = best_cpu;
+					continue;
+				}
 				rcu_read_unlock();
 				return best_cpu;
 			}
@@ -1679,6 +1728,29 @@ static int find_lowest_rq(struct task_struct *task)
 	rcu_read_unlock();
 
 	/*
+	 * If fallback_cpu is valid, all our guesses failed *except* for
+	 * SD_PREFER_SIBLING domain. Now, we can return the fallback cpu.
+	 *
+	 * XXX: Consider the following example, 4 cores SMT2 system:
+	 *
+	 *    LLC [0       -        7]
+	 *    SMT [0 1][2 3][4 5][6 7]
+	 *         o x  o x  x x  x x
+	 *
+	 *    where 'o': occupied and 'x': empty.
+	 *
+	 * A wakeup on cpu0 will exclude cpu1 and choose cpu3, since
+	 * cpu1 is in a SD_PREFER_SIBLING sd and cpu3 is not. However,
+	 * in this case, cpu4 would have been a better choice, since
+	 * cpu3 is a (SMT) thread of an already loaded core.
+	 *
+	 * Doing it 'right' is difficult and expensive. The current
+	 * solution is an acceptable approximation.
+	 */
+	if (fallback_cpu != -1)
+		return fallback_cpu;
+
+	/*
 	 * And finally, if there were no matches within the domains
 	 * just give the caller *something* to work with from the compatible
 	 * locations.
diff --git a/kernel/sched/deadline.c b/kernel/sched/deadline.c
index 0223694..e589576 100644
--- a/kernel/sched/deadline.c
+++ b/kernel/sched/deadline.c
@@ -1319,12 +1319,35 @@ static struct task_struct *pick_earliest_pushable_dl_task(struct rq *rq, int cpu
 
 static DEFINE_PER_CPU(cpumask_var_t, local_cpu_mask_dl);
 
+/*
+ * Find the first cpu in: mask & sd & ~prefer
+ */
+static int find_cpu(const struct cpumask *mask,
+		    const struct sched_domain *sd,
+		    const struct sched_domain *prefer)
+{
+	const struct cpumask *sds = sched_domain_span(sd);
+	const struct cpumask *ps  = prefer ? sched_domain_span(prefer) : NULL;
+	int cpu;
+
+	for_each_cpu(cpu, mask) {
+		if (!cpumask_test_cpu(cpu, sds))
+			continue;
+		if (ps && cpumask_test_cpu(cpu, ps))
+			continue;
+		break;
+	}
+
+	return cpu;
+}
+
 static int find_later_rq(struct task_struct *task)
 {
-	struct sched_domain *sd;
+	struct sched_domain *sd, *prefer = NULL;
 	struct cpumask *later_mask = this_cpu_cpumask_var_ptr(local_cpu_mask_dl);
 	int this_cpu = smp_processor_id();
 	int cpu = task_cpu(task);
+	int fallback_cpu = -1;
 
 	/* Make sure the mask is initialized first */
 	if (unlikely(!later_mask))
@@ -1376,15 +1399,35 @@ static int find_later_rq(struct task_struct *task)
 				return this_cpu;
 			}
 
-			best_cpu = cpumask_first_and(later_mask,
-							sched_domain_span(sd));
 			/*
 			 * Last chance: if a cpu being in both later_mask
 			 * and current sd span is valid, that becomes our
 			 * choice. Of course, the latest possible cpu is
 			 * already under consideration through later_mask.
 			 */
+			best_cpu = find_cpu(later_mask, sd, prefer);
+
 			if (best_cpu < nr_cpu_ids) {
+				/*
+				 * If current domain is SD_PREFER_SIBLING
+				 * flaged, we have to try to check other
+				 * siblings first.
+				 */
+				if (sd->flags & SD_PREFER_SIBLING) {
+					prefer = sd;
+
+					/*
+					 * fallback_cpu should be one
+					 * in the closest domain among
+					 * SD_PREFER_SIBLING domains,
+					 * in case that more than one
+					 * SD_PREFER_SIBLING domains
+					 * exist in the hierachy.
+					 */
+					if (fallback_cpu == -1)
+						fallback_cpu = best_cpu;
+					continue;
+				}
 				rcu_read_unlock();
 				return best_cpu;
 			}
@@ -1393,6 +1436,29 @@ static int find_later_rq(struct task_struct *task)
 	rcu_read_unlock();
 
 	/*
+	 * If fallback_cpu is valid, all our guesses failed *except* for
+	 * SD_PREFER_SIBLING domain. Now, we can return the fallback cpu.
+	 *
+	 * XXX: Consider the following example, 4 cores SMT2 system:
+	 *
+	 *    LLC [0       -        7]
+	 *    SMT [0 1][2 3][4 5][6 7]
+	 *         o x  o x  x x  x x
+	 *
+	 *    where 'o': occupied and 'x': empty.
+	 *
+	 * A wakeup on cpu0 will exclude cpu1 and choose cpu3, since
+	 * cpu1 is in a SD_PREFER_SIBLING sd and cpu3 is not. However,
+	 * in this case, cpu4 would have been a better choice, since
+	 * cpu3 is a (SMT) thread of an already loaded core.
+	 *
+	 * Doing it 'right' is difficult and expensive. The current
+	 * solution is an acceptable approximation.
+	 */
+	if (fallback_cpu != -1)
+		return fallback_cpu;
+
+	/*
 	 * At this point, all our guesses failed, we just return
 	 * 'something', and let the caller sort the things out.
 	 */
@@ -1376,15 +1399,35 @@ static int find_later_rq(struct task_struct *task)
 			/*
-			 * Last chance: if a cpu being in both later_mask
-			 * and current sd span is valid, that becomes our
-			 * choice. Of course, the latest possible cpu is
-			 * already under consideration through later_mask.
+			 * If a cpu being in later_mask & current sd &
+			 * ~prefer sd is valid, that becomes our choice.
+			 * Of course, the latest possible cpu is already
+			 * under consideration through later_mask.
 			 */
-- 
1.9.1

^ permalink raw reply	[flat|nested] 7+ messages in thread

* Re: [PATCH v10 2/2] sched/rt: Add support for SD_PREFER_SIBLING on find_lowest_rq()
  2017-08-24  1:06 ` [PATCH v10 2/2] sched/rt: Add support for SD_PREFER_SIBLING on find_lowest_rq() Byungchul Park
@ 2017-08-24 13:55   ` Steven Rostedt
  2017-08-25  1:31     ` Byungchul Park
  0 siblings, 1 reply; 7+ messages in thread
From: Steven Rostedt @ 2017-08-24 13:55 UTC (permalink / raw)
  To: Byungchul Park
  Cc: peterz, mingo, joel.opensrc, linux-kernel, juri.lelli, kernel-team

On Thu, 24 Aug 2017 10:06:03 +0900
Byungchul Park <byungchul.park@lge.com> wrote:

> Hello Steven,
> 
> Can I keep your reviewed-by with comments modified?

Yes, but I have one nit.


> @@ -1668,9 +1691,35 @@ static int find_lowest_rq(struct task_struct *task)
>  				return this_cpu;
>  			}
>  
> -			best_cpu = cpumask_first_and(lowest_mask,
> -						     sched_domain_span(sd));
> +			/*
> +			 * If a cpu being in lowest_mask & current sd &
> +			 * ~prefer sd is valid, that becomes our choice.

I can't parse the above sentence.

-- Steve

> +			 * Of course, the lowest possible cpu is already
> +			 * under consideration through lowest_mask.
> +			 */
> +			best_cpu = find_cpu(lowest_mask, sd, prefer);
> +

^ permalink raw reply	[flat|nested] 7+ messages in thread

* Re: [PATCH v10 2/2] sched/rt: Add support for SD_PREFER_SIBLING on find_lowest_rq()
  2017-08-24 13:55   ` Steven Rostedt
@ 2017-08-25  1:31     ` Byungchul Park
  2017-08-25  1:47       ` Steven Rostedt
  0 siblings, 1 reply; 7+ messages in thread
From: Byungchul Park @ 2017-08-25  1:31 UTC (permalink / raw)
  To: Steven Rostedt
  Cc: peterz, mingo, joel.opensrc, linux-kernel, juri.lelli, kernel-team

On Thu, Aug 24, 2017 at 09:55:14AM -0400, Steven Rostedt wrote:
> On Thu, 24 Aug 2017 10:06:03 +0900
> Byungchul Park <byungchul.park@lge.com> wrote:
> 
> > Hello Steven,
> > 
> > Can I keep your reviewed-by with comments modified?
> 
> Yes, but I have one nit.
> 
> 
> > @@ -1668,9 +1691,35 @@ static int find_lowest_rq(struct task_struct *task)
> >  				return this_cpu;
> >  			}
> >  
> > -			best_cpu = cpumask_first_and(lowest_mask,
> > -						     sched_domain_span(sd));
> > +			/*
> > +			 * If a cpu being in lowest_mask & current sd &
> > +			 * ~prefer sd is valid, that becomes our choice.
> 
> I can't parse the above sentence.

I might have simplified the comment too much.

~prefer means we want to exclude the domain span having SD_PREFER_SIBLING,
since they should be considered as fallback candidates.

W/o ~prefer, a cpu in the domain might be chosen when we consider higher
domain than the domain having SD_PREFER_SIBLING, as the best cpu, not
fallback.

^ permalink raw reply	[flat|nested] 7+ messages in thread

* Re: [PATCH v10 2/2] sched/rt: Add support for SD_PREFER_SIBLING on find_lowest_rq()
  2017-08-25  1:31     ` Byungchul Park
@ 2017-08-25  1:47       ` Steven Rostedt
  2017-08-25  2:03         ` Byungchul Park
  0 siblings, 1 reply; 7+ messages in thread
From: Steven Rostedt @ 2017-08-25  1:47 UTC (permalink / raw)
  To: Byungchul Park
  Cc: peterz, mingo, joel.opensrc, linux-kernel, juri.lelli, kernel-team

On Fri, 25 Aug 2017 10:31:00 +0900
Byungchul Park <byungchul.park@lge.com> wrote:

> On Thu, Aug 24, 2017 at 09:55:14AM -0400, Steven Rostedt wrote:
> > On Thu, 24 Aug 2017 10:06:03 +0900
> > Byungchul Park <byungchul.park@lge.com> wrote:
> >   
> > > Hello Steven,
> > > 
> > > Can I keep your reviewed-by with comments modified?  
> > 
> > Yes, but I have one nit.
> > 
> >   
> > > @@ -1668,9 +1691,35 @@ static int find_lowest_rq(struct task_struct *task)
> > >  				return this_cpu;
> > >  			}
> > >  
> > > -			best_cpu = cpumask_first_and(lowest_mask,
> > > -						     sched_domain_span(sd));
> > > +			/*
> > > +			 * If a cpu being in lowest_mask & current sd &
> > > +			 * ~prefer sd is valid, that becomes our choice.  
> > 
> > I can't parse the above sentence.  
> 
> I might have simplified the comment too much.
> 
> ~prefer means we want to exclude the domain span having SD_PREFER_SIBLING,
> since they should be considered as fallback candidates.
> 
> W/o ~prefer, a cpu in the domain might be chosen when we consider higher
> domain than the domain having SD_PREFER_SIBLING, as the best cpu, not
> fallback.

That actually wasn't the hard part. It was the entire sentence that is
very confusing. What is the definition of "is valid" in this context?
Do you just mean "< nr_cpu_ids"? If that's the case you are also being
redundant. Because if the "if" statement is true, then it will be
valid. What about just saying something like:


/*
 * If a cpu exists that is in the lowest_mask and the current sd span,
 * but not in the prefer sd span, then that becomes our choice.
 */

-- Steve

^ permalink raw reply	[flat|nested] 7+ messages in thread

* Re: [PATCH v10 2/2] sched/rt: Add support for SD_PREFER_SIBLING on find_lowest_rq()
  2017-08-25  1:47       ` Steven Rostedt
@ 2017-08-25  2:03         ` Byungchul Park
  0 siblings, 0 replies; 7+ messages in thread
From: Byungchul Park @ 2017-08-25  2:03 UTC (permalink / raw)
  To: Steven Rostedt
  Cc: peterz, mingo, joel.opensrc, linux-kernel, juri.lelli, kernel-team

On Thu, Aug 24, 2017 at 09:47:59PM -0400, Steven Rostedt wrote:
> /*
>  * If a cpu exists that is in the lowest_mask and the current sd span,
>  * but not in the prefer sd span, then that becomes our choice.
>  */

I like it. It looks much better to me.

I will change it.

Thank you.

^ permalink raw reply	[flat|nested] 7+ messages in thread

end of thread, other threads:[~2017-08-25  2:03 UTC | newest]

Thread overview: 7+ messages (download: mbox.gz / follow: Atom feed)
-- links below jump to the message on this page --
2017-08-24  1:06 [PATCH v10 0/2] Make find_later_rq() choose a closer cpu in topology Byungchul Park
2017-08-24  1:06 ` [PATCH v10 1/2] sched/deadline: Add support for SD_PREFER_SIBLING on find_later_rq() Byungchul Park
2017-08-24  1:06 ` [PATCH v10 2/2] sched/rt: Add support for SD_PREFER_SIBLING on find_lowest_rq() Byungchul Park
2017-08-24 13:55   ` Steven Rostedt
2017-08-25  1:31     ` Byungchul Park
2017-08-25  1:47       ` Steven Rostedt
2017-08-25  2:03         ` Byungchul Park

This is a public inbox, see mirroring instructions
for how to clone and mirror all data and code used for this inbox

all inboxes | Powered by JetHome®